from enum import Enum
from distutils.version import StrictVersion
from functools import total_ordering
import logging

from pydotfiles.utils.io import run_command

logger = logging.getLogger(__name__)


class Setting:

    def __init__(self, name, valid_version_range, command, enabled=True, description=None, check_command=None, expected_check_state=None, run_as_sudo=False, check_output=True):
        self.name = name
        self.enabled = enabled
        self.description = description
        self.valid_version_range = valid_version_range
        self.command = command
        self.check_command = check_command
        self.expected_check_state = expected_check_state
        self.run_as_sudo = run_as_sudo
        self.check_output = check_output

    def __str__(self):
        return f"Setting(name={self.name}, enabled={self.enabled}, description={self.description}, valid_version_range={self.valid_version_range}, command={self.command}, check_command={self.check_command}, expected_check_state={self.expected_check_state})"

    def should_run(self, current_version, sudo_password):
        if not self.enabled:
            return False

        if not self.valid_version_range.is_in_range(current_version):
            return False

        if self.check_command is None:
            return True

        current_status_check_result = run_command(self.check_command, self.run_as_sudo, sudo_password, check_output=self.check_output)

        if current_status_check_result != self.expected_check_state:
            logger.info(f"Setting: Expected value not found, running command now [action={self.name}, expected={self.expected_check_state}, found={current_status_check_result}, run_as_sudo={self.run_as_sudo}")

        return current_status_check_result != self.expected_check_state

    def run(self, sudo_password):
        run_command(self.command, self.run_as_sudo, sudo_password, check_output=self.check_output)


class VersionRange:

    def __init__(self, start=None, end=None):
        """
        A None start is taken to be the equivalent
        of "since the beginning of time"

        A None end is taken to be the equivalent of
        "currently supported"
        """

        if start is not None and not isinstance(start, StrictVersion) and not isinstance(start, MacVersion):
            raise ValueError(f"Unable to create a version range, passed in type for start was invalid [start={type(start)}], should have been a StrictVersion")

        if end is not None and not isinstance(end, StrictVersion) and not isinstance(start, MacVersion):
            raise ValueError(f"Unable to create a version range, passed in type for start was invalid [end={type(start)}], should have been a StrictVersion")

        self.start = start
        self.end = end

    def __str__(self):
        return f"VersionRange(start={self.start}, end={self.end})"

    def is_in_range(self, current_version):
        # Infinite range case
        if self.start is None and self.end is None:
            return True

        if self.start is None and self.end is not None:
            return current_version <= self.end

        if self.end is None and self.start is not None:
            return self.start <= current_version

        # For somebody who likes to keep their books clean
        return self.start <= current_version <= self.end


@total_ordering
class MacVersion(Enum):

    YOSEMITE = StrictVersion("10.10")

    EL_CAPITAN = StrictVersion("10.11")

    SIERRA = StrictVersion("10.12")

    HIGH_SIERRA = StrictVersion("10.13")

    MOJAVE = StrictVersion("10.14")

    @staticmethod
    def from_version(version):
        if version is None:
            raise ValueError("MacVersion: Can't identify none version number")

        if isinstance(version, str):
            version = StrictVersion(version)

        major_version = version.version[0]
        minor_version = version.version[1]

        truncated_version = StrictVersion(f"{major_version}.{minor_version}")
        return MacVersion(truncated_version)

    @staticmethod
    def from_name(name):
        if name is None:
            raise KeyError("MacVersion: Can't identify none version name")
        return MacVersion[name.upper()]

    def __lt__(self, other):
        return self.value < other.value



import logging
import os
import json
import yaml
import jsonschema
from pathlib import Path
from pkg_resources import resource_stream, resource_filename

from .utils import set_logging, load_data_from_file
from .exceptions import ValidationError, ValidationErrorReason

logger = logging.getLogger(__name__)


class ConfigMapper:
    """
    Maps a given configuration file with a set
    schema to an internal configuration data
    structure
    """

    @staticmethod
    def get_schema(version, schema):
        schema_file = resource_stream(f"pydotfiles.resources.schemas.{version}", f"{schema}.json")
        return json.load(schema_file)

    @staticmethod
    def get_resolver(version, schema):
        schema_file_name = resource_filename(f"pydotfiles.resources.schemas.{version}", f"{schema}.json")
        schema_file_path = Path(schema_file_name)
        return jsonschema.RefResolver(f"file://{str(schema_file_path.parent)}/", None)


class Validator:
    """
    Validates that a given directory, module,
    or file is pydotfiles-compliant
    """

    def __init__(self, is_quiet=False, is_verbose=False):
        self.is_quiet = is_quiet
        self.is_verbose = is_verbose
        set_logging(is_quiet, is_verbose)

    def validate_directory(self, directory):
        if directory is None:
            raise ValidationError(ValidationErrorReason.INVALID_TARGET, "The passed in directory is invalid [directory=None]")

        # Sanity check: Does the directory exist
        if not os.path.isdir(directory):
            raise ValidationError(ValidationErrorReason.INVALID_TARGET, f"The passed in directory is invalid [directory={directory}]")

        # Sanity check: Converts to a Path object if not one already
        if type(directory) is str:
            directory = Path(directory)

        logger.info(f"Validator: Validating directory [directory={directory}]")

        # Generates a set of files that we need to validate from a tree structure
        initial_files_to_validate = set()
        for path_prefix, directory_names, file_names in os.walk(directory):
            for file_name_path in file_names:
                file_name = str(file_name_path)
                if file_name.endswith(".json") or file_name.endswith(".yaml") or file_name.endswith(".yml"):
                    initial_files_to_validate.add(os.path.join(path_prefix, file_name))

        validation_exceptions = []
        for file_to_validate in initial_files_to_validate:
            try:
                self.validate_file(file_to_validate)
            except ValidationError as e:
                if self.is_verbose:
                    logger.exception(e.help_message)
                validation_exceptions.append(e)

        number_of_validation_errors = len(validation_exceptions)
        if number_of_validation_errors == 0:
            logger.info(f"Validator: Successfully validated directory [directory={directory}]")
        else:
            logger.error(f"Validator: Directory failed validation [directory={directory}, number_of_validation_errors={number_of_validation_errors}]")
            raise validation_exceptions[0]

    def validate_file(self, file):
        if file is None:
            raise ValidationError(ValidationErrorReason.INVALID_TARGET, "Validator: No file was passed in")

        # Sanity check: Does the file exist
        if not os.path.isfile(file):
            validation_error = ValidationError(ValidationErrorReason.INVALID_TARGET, f"Validator: The passed in file is invalid")
            validation_error.context_map['file'] = file
            raise validation_error

        # Sanity check: Converts to a Path object if not one already
        if type(file) is str:
            file = Path(file)

        logger.info(f"Validator: Validating file [file={file}]")

        # Validates the format
        try:
            file_data = load_data_from_file(file)
        except json.JSONDecodeError as e:
            validation_error = ValidationError(ValidationErrorReason.INVALID_SYNTAX, f"Validator: An invalid JSON syntax error was detected")
            validation_error.context_map['file'] = file
            raise validation_error from e
        except yaml.YAMLError as e:
            validation_error = ValidationError(ValidationErrorReason.INVALID_SYNTAX, f"Validator: An invalid YAML syntax error was detected")
            validation_error.context_map['file'] = file
            raise validation_error from e

        # Validates the schema
        try:
            self.validate_data(file_data)
        except ValidationError as e:
            e.context_map['file_name'] = file
            raise e

        # Recursively dispatches to validate other files if needed
        if file_data.get('schema') == 'core':
            defaults_setting_file_name = file_data.get('os', {}).get('default_settings_file')
            if defaults_setting_file_name is not None:
                defaults_settings_file_path = Path.joinpath(file.parent, defaults_setting_file_name)
                self.validate_file(defaults_settings_file_path)

        logger.info(f"Validator: Successfully validated file [file={file}]")

    @staticmethod
    def validate_data(data):
        if data is None:
            raise ValidationError(ValidationErrorReason.INVALID_DATA, "Validator: No parsed config data was returned")

        # Isolates for which version we need to get
        version = data.get("version")
        if version is None:
            raise ValidationError(ValidationErrorReason.INVALID_SCHEMA_VERSION, "Validator: The schema version was not found (is there a 'version' field?)")

        # Isolates for which schema type we need to get
        schema_type = data.get("schema")
        if schema_type is None:
            raise ValidationError(ValidationErrorReason.INVALID_SCHEMA_TYPE, "Validator: The schema type was not found (is there a 'schema' field?)")

        # Retrieves the required schema
        schema = ConfigMapper.get_schema(version, schema_type)

        # We need a custom resolver since we're referencing other schemas
        # For more information, see https://stackoverflow.com/a/53968771
        resolver = ConfigMapper.get_resolver(version, schema_type)

        # Validates the given data to the schema
        try:
            jsonschema.validate(data, schema, resolver=resolver)
        except jsonschema.exceptions.ValidationError as e:
            validator_error = ValidationError(ValidationErrorReason.INVALID_SCHEMA)
            validator_error.context_map['reason'] = e.message
            raise validator_error from e



from .enums import PydotfilesErrorReason, ValidationErrorReason


class PydotfilesError(Exception):

    def __init__(self, reason, help_message_override=None):
        super().__init__()
        self.reason = reason
        self.help_message_override = help_message_override

    @property
    def help_message(self):
        return PydotfilesErrorReason.get_help_message(self.reason) if self.help_message_override is None else self.help_message_override


class ValidationError(Exception):

    def __init__(self, reason, help_message_override=None, context_map=None):
        super().__init__()
        self.reason = reason
        self.help_message_override = help_message_override
        if context_map is None:
            context_map = {
                "help": "If you need additional context like a stack trace, please run in verbose mode (with the -v flag)"
            }
        self.context_map = context_map

    @property
    def help_message(self):
        undecorated_original_help_message = ValidationErrorReason.get_help_message(self.reason) if self.help_message_override is None else self.help_message_override

        if len(self.context_map) == 0:
            serialized_context_decoration = ""
        else:
            serialized_context_decoration = '\n'.join([f"\t{context_type}: {context_reason}" for context_type, context_reason in self.context_map.items()])

        decorated_error_message = f"{undecorated_original_help_message} [\n{serialized_context_decoration}\n]"
        return decorated_error_message



# General imports
from json import loads as json_loads
from json import dumps as json_dumps
from logging import getLogger
from pathlib import Path
try:
    from importlib.resources import files
except ImportError:
    from importlib_resources import files

# 3rd party imports
from jsonschema import RefResolver
from jsonschema import Draft202012Validator
from jsonschema.exceptions import ValidationError

logger = getLogger(__name__)


class AlphaValidator:
    def __init__(self):
        self.schema_store: dict[str, dict] = self.__load_schema_store()
        logger.debug(f"The schema store looks like:\n-----------------\n{json_dumps(self.schema_store, indent=4, sort_keys=True)}\n---------------")
        self.resolver = RefResolver("", "", self.schema_store)

    def validate(self, file_path: Path, data: dict) -> bool:
        if data is None:
            return False

        # Isolates for which version we need to get
        version: str = data.get("version")
        if version is None:
            return False

        # Isolates for which schema type we need to get
        schema_type: str = data.get("schema")
        if schema_type is None:
            return False

        identified_schema = self.schema_store.get(f"{version}.{data.get('schema')}.json")
        logger.debug(f"The identified schema to use in validation is:\n{json_dumps(identified_schema, indent=4, sort_keys=True)}\n")

        validator = Draft202012Validator(identified_schema, resolver=self.resolver)
        try:
            validator.validate(data)
            logger.debug(f"Validation: Successfully validated file path: {file_path} using schema: pydotfiles.resources.schemas.{version}.{schema_type}")
            return True
        except ValidationError:
            logger.exception(f"Error when attempting to validate configuration file:\n{data=}\n")
            return False

    def __load_schema_store(self) -> dict[str, dict]:
        """
        Hardcoded for now to avoid doing a path walk and because this shouldn't change often
        """
        common_schema = self.__get_schema("alpha", "common")
        core_schema = self.__get_schema("alpha", "core")
        default_settings_schema = self.__get_schema("alpha", "default_settings")
        developer_environments_schema = self.__get_schema("alpha", "developer_environments")
        manifest_schema = self.__get_schema("alpha", "manifest")
        return {
            "alpha.common.json": common_schema,
            "alpha.core.json": core_schema,
            "alpha.default_settings.json": default_settings_schema,
            "alpha.developer_environments.json": developer_environments_schema,
            "alpha.manifest.json": manifest_schema,
        }

    def __get_schema(self, version: str, schema: str) -> dict:
        my_resources = files("pydotfiles")
        data = (my_resources / "resources" / "schemas" / version / f"{schema}.json").read_text()
        return json_loads(data)



# General imports
from dataclasses import dataclass
from enum import Enum
from typing import Optional, Dict


class DevLanguage(Enum):
    python = "python"
    ruby = "ruby"
    golang = "golang"
    java = "java"
    rust = "rust"


@dataclass(frozen=True, order=True)
class VirtualEnvironment:
    version: str
    name: str

    @staticmethod
    def from_dict(data: Dict) -> 'VirtualEnvironment':
        return VirtualEnvironment(
            version=data.get("version"),
            name=data.get("name")
        )


@dataclass(frozen=True, order=True)
class Plugin:
    name: str
    virtual_environments: list[VirtualEnvironment]

    @staticmethod
    def from_dict(data: Dict) -> 'Plugin':
        virtual_environments = [VirtualEnvironment.from_dict(virtual_environment) for virtual_environment in data.get("virtual_environments")]
        return Plugin(
            name=data.get("name"),
            virtual_environments=virtual_environments
        )


@dataclass(frozen=True, order=True)
class EnvironmentManager:
    name: str
    plugins: list[Plugin]

    @staticmethod
    def from_dict(data: Optional[Dict]) -> Optional['EnvironmentManager']:
        if data is None:
            return None

        plugins = [] if data.get("plugins") is None else [Plugin.from_dict(plugin) for plugin in data.get("plugins")]
        return EnvironmentManager(
            name=data.get("name"),
            plugins=plugins
        )


@dataclass(frozen=True, order=True)
class Environment:
    language: DevLanguage
    versions: list[str]
    environment_manager: Optional[EnvironmentManager]

    @staticmethod
    def from_dict(data: list[Dict]) -> list['Environment']:
        return [Environment.from_environment_dict(datum) for datum in data]

    @staticmethod
    def from_environment_dict(data: Dict) -> 'Environment':
        return Environment(
            language=DevLanguage[data.get("language")],
            versions=data.get("versions"),
            environment_manager=EnvironmentManager.from_dict(data.get("environment_manager"))
        )



# General imports
from pathlib import Path
from os import access as os_access
from os import X_OK as IS_EXECUTABLE
from hashlib import sha256 as hashlib_sha256
from logging import getLogger

# Project imports
from pydotfiles.v4.builder.environment import template_dev_env_manager_installation, \
    template_dev_env_manager_plugins_installation, template_dev_env_manager_version_installation
from pydotfiles.v4.common.alpha import Environment, OSName
from .common import Symlink

logger = getLogger(__name__)

##
# Templating functions for helper functions
##


def template_starter_helper_functions() -> str:
    return r"""
    function info () {
        printf "\r  [ \033[00;34m..\033[0m ] %s\n" "$1"
    }

    function user () {
        printf "\r  [ \033[0;33m??\033[0m ] %s " "$1"
    }

    function success () {
        printf "\r\033[2K  [ \033[00;32mOK\033[0m ] %s\n" "$1"
    }

    function warn () {
        printf "\r\033[2K  [\033[0;31mWARN\033[0m] %s\n" "$1"
    }

    function fail () {
        printf "\r\033[2K  [\033[0;31mFAIL\033[0m] %s\n" "$1"
        echo ''
        exit 1
    }
    """

##
# Templating function for dev environment
##


def template_dev_environment_setup(data: Environment, active_os: OSName) -> str:
    dev_environment_installation = ""

    dev_environment_installation += template_dev_env_manager_installation(data, active_os)
    dev_environment_installation += template_dev_env_manager_version_installation(data)
    dev_environment_installation += template_dev_env_manager_plugins_installation(data)

    return dev_environment_installation

##
# Templating functions for actions
##


def template_copy_file(origin: Path, destination: Path) -> str:
    """
    Returns the equivalent bash command in order to copy a file.
    NOTE: origin is a _relative_ path from the root of the dotfiles directory
    NOTE: destination is an _absolute_ path
    NOTE: We removed sudo ability here, but can re-add it in if necessary
    """

    return f"""
    if [[ -f {destination} ]]; then
        info "Copy: File already exists in {destination}"
    else
        if mkdir -p {destination.parent} && cp {origin} {destination} ; then
            success "Copy: Successfully copied file {origin} to {destination}"
        else
            fail "Copy: Failed to copy file {origin} to {destination}"
        fi
    fi
    """


def template_symlink_file(symlink: Symlink) -> str:
    """
    Returns the equivalent bash command in order to symlink a file
    NOTE: origin is a _relative_ path from the root of the dotfiles directory
    NOTE: destination is an _absolute_ path
    """
    origin = symlink.script_origin_file
    destination = symlink.script_destination_file

    return f"""
    if [[ -L {destination} ]]; then
        info "Symlink: File already exists in {destination}"
    else
        if {'sudo ' if symlink.is_sudo else ''}mkdir -p {destination.parent} && {'sudo ' if symlink.is_sudo else ''}ln -s {'' if symlink.initial_origin_file is None else '$(pwd)/'}{origin} {destination} ; then
            success "Symlink: Successfully symlinked {origin} to {destination}"
        else
            fail "Symlink: Failed to symlink {origin} to {destination}"
        fi
    fi
    """


def template_run_script(script: Path, use_sudo: bool) -> str:
    """
    Returns the equivalent bash command in order to copy a file
    """
    return f"""
    if {'sudo ' if use_sudo else ''}{script}; then
        success "Script file: {script} executed successfully"
    else
        fail "Script file: {script} failed to executed"
    fi
    """


##
# Utility functions
##


def is_moved(origin: Path, destination: Path) -> bool:
    # Enables fast-failing based on existence
    if not destination.is_file():
        return False

    if origin.is_file():
        return False

    return True


def is_broken_link(file: Path) -> bool:
    return file.is_symlink() and not file.exists()


def is_linked(origin: Path, destination: Path) -> bool:
    return destination.is_symlink() and destination.resolve() == origin.resolve()


def is_copied(origin: Path, destination: Path) -> bool:
    # If the file already exists then no need to do anything
    if destination.is_file():
        return True

    # Enables fast-failing based on type
    if destination.is_symlink():
        return False

    # Enables fast-failing based on file size

    origin_file_size = origin.stat().st_size
    destination_file_size = destination.stat().st_size

    if origin_file_size != destination_file_size:
        return False

    # Enables fast-failing based on metadata
    origin_last_modified_time = origin.stat().st_mtime
    destination_last_modified_time = destination.stat().st_mtime

    if origin_last_modified_time != destination_last_modified_time:
        return False

    # Loads in the files and checks that their hashes are the same
    origin_file_hash = hash_file(origin)
    destination_file_hash = hash_file(destination)
    return origin_file_hash == destination_file_hash


def is_executable(file: Path) -> bool:
    return os_access(file, IS_EXECUTABLE)


def hash_file(file_path: Path, block_size=65536) -> str:
    """
    Hashes a given file and returns the digest in hexadecimal
    """
    hasher = hashlib_sha256()
    with file_path.open('rb') as raw_file:
        buf = raw_file.read(block_size)
        while len(buf) > 0:
            hasher.update(buf)
            buf = raw_file.read(block_size)

    return hasher.hexdigest()



# General imports
from logging import getLogger

# Project imports
from pydotfiles.v4.common.alpha import Environment, OSName
from pydotfiles.v4.common.alpha.environment import Plugin

logger = getLogger(__name__)

##
# Developer environment manager installation
##


def template_dev_env_manager_installation(data: Environment, active_os: OSName) -> str:
    environment_managers = {
        'pyenv': get_pyenv_installation_template,
        'jenv': get_jenv_installation_template
    }
    return environment_managers.get(data.environment_manager.name, lambda: "")(active_os)


def get_pyenv_installation_template(active_os: OSName) -> str:
    pyenv_dispatcher = {
        OSName.linux: get_pyenv_linux_installation_template,
        OSName.macos: get_pyenv_macos_installation_template,
    }

    return pyenv_dispatcher.get(active_os)()


def get_pyenv_linux_installation_template() -> str:
    return r"""

    ##
    # Installs pyenv if it's not yet installed
    ##
    function linux_setup() {
        linux_install_pyenv_dependencies
        linux_install_pyenv
        linux_install_pyenv_environment
    }
    
    function linux_install_pyenv_dependencies() {
        # Grabbed from https://www.liquidweb.com/kb/how-to-install-pyenv-on-ubuntu-18-04/
        apt update -y
        apt install -y make build-essential libssl-dev zlib1g-dev \
        libbz2-dev libreadline-dev libsqlite3-dev wget curl llvm libncurses5-dev \
        libncursesw5-dev xz-utils tk-dev libffi-dev liblzma-dev python-openssl \
        git
    }
    
    function linux_install_pyenv() {
        git clone https://github.com/pyenv/pyenv.git ~/.pyenv
    }
    
    function linux_install_pyenv_environment() {
        info "Pyenv Environment: Checking for correct pyenv environment variable setup"
        if [[ -z "${PYENV_ROOT}" ]]; then
            info "Pyenv Environment: Missing pyenv environment variables"
            {
                echo "# Pydotfiles: Automatically set pyenv settings"
                echo "export PYENV_ROOT=$HOME/.pyenv"
                echo "if which pyenv > /dev/null; then eval \"\$(pyenv init -)\"; fi"
                echo "if which pyenv-virtualenv-init > /dev/null; then eval \"\$(pyenv virtualenv-init -)\"; fi"
            } >> ~/.bash_profile

            # shellcheck disable=SC1090
            source "${HOME}/.bash_profile"
            success "Pyenv Environment: Successfully added in pyenv environment variables"
        else
            success "Pyenv Environment: Already have set pyenv environment variables"
        fi
    }
    linux_setup
    """


def get_pyenv_macos_installation_template() -> str:
    return r"""

    ##
    # Installs pyenv if it's not yet installed
    ##

    function mac_setup() {
        mac_install_xcode
        mac_install_homebrew
        mac_install_pyenv
        mac_install_pyenv_environment
    }

    function mac_install_xcode() {
        # Installs xcode command line tools and accepts the license
        if xcode-select -p &> /dev/null ; then
            success "Xcode: MacOS command-line tools are already installed"
        else
            info "Xcode: MacOS command-line tools were not installed, installing now"
            if install_xcode ; then
                success "Xcode: Successfully installed macOS command-line tools"
            else
                fail "Xcode: Failed to install macOS command-line tools"
            fi
        fi
    }

    function mac_install_homebrew() {
        info "Package Manager: Checking for the existance of Homebrew"
        if [[ $(command -v brew) == "" ]]; then
            info "Package Manager: Homebrew was not found, installing now"

            # shellcheck disable=SC2028,SC1117
            if echo "\r" | /usr/bin/ruby -e "$(curl -fsSL https://raw.githubusercontent.com/Homebrew/install/master/install)" ; then
                success "Package Manager: Successfully installed Homebrew"
            else
                fail "Package Manager: Failed to install Homebrew"
            fi
        else
            success "Package Manager: Homebrew is already installed"
        fi
    }

    function mac_install_pyenv() {
        info "Pyenv: Checking for pyenv installation"
        if [[ $(command -v pyenv | grep pyenv) == "" ]]; then
            info "Pyenv: Missing pyenv package, installing now"
            if brew install pyenv ; then
                success "Pyenv: Successfully installed pyenv"
            else
                fail "Pyenv: Failed to install pyenv"
            fi
        else
            success "Pyenv: Pyenv is already installed"
        fi
    }

    function mac_install_pyenv_environment() {
        info "Pyenv Environment: Checking for correct pyenv environment variable setup"
            if [[ -z "${PYENV_ROOT}" ]]; then
            info "Pyenv Environment: Missing pyenv environment variables"
            {
                echo "# Pydotfiles: Automatically set pyenv settings"
                echo "export PYENV_ROOT=$HOME/.pyenv"
                echo "if which pyenv > /dev/null; then eval \"\$(pyenv init -)\"; fi"
                echo "if which pyenv-virtualenv-init > /dev/null; then eval \"\$(pyenv virtualenv-init -)\"; fi"
            } >> ~/.bash_profile

            # shellcheck disable=SC1090
            source "${HOME}/.bash_profile"
            success "Pyenv Environment: Successfully added in pyenv environment variables"
        else
            success "Pyenv Environment: Already have set pyenv environment variables"
        fi
    }
    mac_setup

    """


def get_jenv_installation_template(active_os: OSName) -> str:
    jenv_installer_dispatcher = {
        OSName.linux: get_jenv_linux_installation_template,
        OSName.macos: get_jenv_macos_installation_template
    }
    return jenv_installer_dispatcher.get(active_os)()


def get_jenv_linux_installation_template() -> str:
    return r"""
    git clone https://github.com/jenv/jenv.git ~/.jenv
    """


def get_jenv_macos_installation_template() -> str:
    return r"""
    # brew install jenv
    """

##
# Developer environment manager plugin installation
##


def template_dev_env_manager_plugins_installation(data: Environment) -> str:
    environment_manager_plugins = {
        'pyenv-virtualenv': get_pyenv_virtualenv_installation_template,
    }

    plugins_template = ""
    for plugin in data.environment_manager.plugins:
        plugins_template += environment_manager_plugins.get(plugin.name, lambda: "")()
        plugins_template += environment_manager_plugin_setup(plugin)
    return plugins_template


def get_pyenv_virtualenv_installation_template() -> str:
    return r"""
    function pyenv_install_pyenv_virtualenv() {
        info "Pyenv-virtualenv: Checking for pyenv-virtualenv installation"
        if [[ $(brew list | grep "pyenv-virtualenv") == "" ]]; then
            info "Pyenv-virtualenv: Missing pyenv-virtualenv package, installing now"
            if brew install pyenv-virtualenv ; then
                success "Pyenv-virtualenv: Successfully installed pyenv-virtualenv"
            else
                fail "Pyenv-virtualenv: Failed to install pyenv-virtualenv"
            fi
        else
            success "Pyenv-virtualenv: Pyenv-virtualenv is already installed"
        fi
    }
    pyenv_install_pyenv_virtualenv
    """


def environment_manager_plugin_setup(data: Plugin) -> str:
    environment_manager_plugin_setup_methods = {
        'pyenv-virtualenv': get_pyenv_virtualenv_virtual_env_setup
    }
    return environment_manager_plugin_setup_methods.get(data.name, lambda x: "")(data)


def get_pyenv_virtualenv_virtual_env_setup(data: Plugin) -> str:
    versions_to_install = f"python_base_versions=({' '.join([virtual_env.version for virtual_env in data.virtual_environments])})\n"
    virtual_envs_to_install = f"python_virtual_env_aliases=({' '.join([virtual_env.name for virtual_env in data.virtual_environments])})\n"
    driver_function = r"""
    function pyenv_install_virtual_envs() {
        info "Pyenv Virtualenv: Checking that all pyenv virtual environments are installed"

        # We choose not to use an associative array in case bash versions used are below Bash 4,
        # which is when associative arrays were introduced
        for index in "${!python_base_versions[@]}"; do
            pyenv_install_single_virtualenv "${python_base_versions[$index]}" "${python_virtual_env_aliases[$index]}"
        done
        success "Pyenv Virtualenv: All pyenv global environments are installed"
    }

    function pyenv_install_single_virtualenv() {
        version=$1
        alias_name=$2

        if [[ $(pyenv versions | grep "${alias_name}") == "" ]]; then
            info "Pyenv Virtualenv: The virtual env '${alias_name}' was not found, installing now"
            if pyenv virtualenv "${version}" "${alias_name}" &> /dev/null ; then
                success "Pyenv Virtualenv: Successfully installed the virtual env '${alias_name}'"
            else
                fail "Pyenv Virtualenv: Failed to install the virtual env '${alias_name}'"
            fi
        else
            success "Pyenv Virtualenv: The virtual env '${alias_name}' is already installed"
        fi
    }
    pyenv_install_virtual_envs
    """
    return versions_to_install + virtual_envs_to_install + driver_function

##
# Developer environment manager version installation
##


def template_dev_env_manager_version_installation(data: Environment) -> str:
    environment_manager_version_installer = {
        'pyenv': get_pyenv_version_install_template
    }
    return environment_manager_version_installer.get(data.environment_manager.name, lambda x: "")(data)


def get_pyenv_version_install_template(data: Environment) -> str:
    versions_to_install = f"python_base_versions=({' '.join(data.versions)})\n"
    base_installation_driver = r"""
    function pyenv_install_base_environments() {
        info "Pyenv Base Environment: Checking that all pyenv base environments are installed"
        for version in "${python_base_versions[@]}"; do
            pyenv_install_single_base_environment "${version}"
        done
        success "Pyenv Base Environment: All pyenv base environments look good"
    }
    pyenv_install_base_environments
    """

    pyenv_installation_functions = r"""
    function pyenv_install_single_base_environment() {
        version=$1
        if [[ $(pyenv versions | grep "${version}") == "" ]]; then
            info "Pyenv: Python version ${version} is not installed yet, installing now"
    
            # Checks if pyenv has the correct version first
            info "Pyenv: Checking for python version ${version}"
            if [[ $(pyenv install --list | grep "${version}") == "" ]]; then
                fail "Pyenv: Python version ${version} was not found, please check it is valid and pyenv is up to date"
            else
                success "Pyenv: Python version ${version} was found"
            fi
    
            # Runs through normal install if possible, using CFlag if not
            info "Pyenv: Attempting normal install (might take a few minutes)"
            if pyenv install "${version}" &> /dev/null ; then
                success "Pyenv: Python version ${version} is now installed"
            else
                warn "Pyenv: Python version ${version} failed to install, attempting to run with CFlags"
                pyenv_install_base_environment_fallback "${version}"
            fi
        else
            success "Pyenv: Python version ${version} is already installed"
        fi
    }
    
    function pyenv_install_base_environment_fallback() {
        version=$1
    
        if CFLAGS="-I$(xcrun --show-sdk-path)/usr/include" pyenv install "${version}" &> /dev/null ; then
            success "Pyenv: Python version ${version} is now installed"
        else
            warn "Pyenv: CFlag setting when installing ${version} failed, attempting to install with CFlags and LDFlags"
            if CFLAGS="-I$(brew --prefix openssl)/include" LDFLAGS="-L$(brew --prefix openssl)/lib" pyenv install "${version}" &> /dev/null ; then
                success "Pyenv: Python version ${version} is now installed"
            else
                fail "Pyenv: Python version ${version} failed to install"
            fi
        fi
    }
    """

    return versions_to_install + pyenv_installation_functions + base_installation_driver



# General imports
from logging import getLogger
from logging import StreamHandler, Formatter, Filter
from logging import WARNING, DEBUG, INFO, ERROR, FATAL
from logging import root as root_logger

from sys import stderr, stdout
from sys import exit as sys_exit

##
# Public method
##

logger = getLogger(__name__)


def set_logging(is_quiet: bool, is_verbose: bool) -> None:
    if is_quiet:
        logging_level = WARNING
    elif is_verbose:
        logging_level = DEBUG
    else:
        logging_level = INFO

    # Enables normal logging to stdout and errors to stderr
    stderr_handler = StreamHandler(stderr)
    stderr_handler.setLevel(WARNING)
    stderr_handler.setFormatter(PrettyLogFormatter())
    root_logger.addHandler(stderr_handler)

    # See https://github.com/Robpol86/sphinxcontrib-versioning/blob/master/sphinxcontrib/versioning/setup_logging.py
    stdout_handler = StreamHandler(stdout)
    stdout_handler.setLevel(logging_level)
    stdout_handler.setFormatter(PrettyLogFormatter())
    stdout_handler.addFilter(type('', (Filter,), {'filter': staticmethod(lambda r: r.levelno < WARNING)}))
    root_logger.addHandler(stdout_handler)

    root_logger.setLevel(logging_level)

##
# Helper classes and methods
##


class PrettyLogFormatter(Formatter):
    """
    Presentation-layer pretty printing, enabling
    clear guides to an end-user via ANSI escape
    sequences. Should be attached to a logger.

    For more information see:
    https://stackoverflow.com/questions/1343227/can-pythons-logging-format-be-modified-depending-on-the-message-log-level/27835318#27835318
    """

    DEBUG_BLUE = "\033[94m"
    INFO_GREEN = "\033[92m"
    WARNING_YELLOW = "\033[0;33m"
    ERROR_ORANGE = "\033[38:2:255:165:0m"
    FATAL_RED = "\033[31m"

    END_COLOUR = '\033[0m'

    def __init__(self):
        super().__init__()
        self.formatters = {
            DEBUG: Formatter(f"%(asctime)s [{self.DEBUG_BLUE}DEBUG::%(name)s{self.END_COLOUR}]: %(message)s", datefmt=self.date_format),
            INFO: Formatter(f"%(asctime)s [{self.INFO_GREEN}%(levelname)s::%(name)s{self.END_COLOUR}]: %(message)s", datefmt=self.date_format),
            WARNING: Formatter(f"%(asctime)s [{self.WARNING_YELLOW}WARN::%(name)s{self.END_COLOUR}]: %(message)s", datefmt=self.date_format),
            ERROR: Formatter(f"%(asctime)s [{self.ERROR_ORANGE}ERROR::%(name)s{self.END_COLOUR}]: %(message)s", datefmt=self.date_format),
            FATAL: Formatter(f"%(asctime)s [{self.FATAL_RED}FATAL::%(name)s{self.END_COLOUR}]: %(message)s", datefmt=self.date_format),
        }

        self.default_format = Formatter("%(levelname)s: %(name)s: %(message)s")

    @property
    def date_format(self):
        return "%Y-%m-%d %I:%M:%S %p"

    def format(self, record):
        return self.formatters.get(record.levelno, self.default_format).format(record)


class PrettyPrint:
    """
    Presentation-layer pretty printing, enabling
    clear guides to an end-user via ANSI escape
    sequences
    """

    SUCCESS_GREEN = '\033[92m'
    INFO_BLUE = '\033[94m'
    USER_YELLOW = '\033[0;33m'
    WARN_ORANGE = '\033[38:2:255:165:0m'
    FAIL_RED = '\033[31m'

    # Other settings not turned on yet
    # HEADER = '\033[95m'
    # BOLD = '\033[1m'
    # UNDERLINE = '\033[4m'

    END_COLOUR = '\033[0m'

    @staticmethod
    def success(message):
        print(f"[ {PrettyPrint.SUCCESS_GREEN}OK{PrettyPrint.END_COLOUR} ] {message}")

    @staticmethod
    def info(message):
        print(f"[{PrettyPrint.INFO_BLUE}INFO{PrettyPrint.END_COLOUR}] {message}")

    @staticmethod
    def user(message):
        print(f"[{PrettyPrint.USER_YELLOW}USER{PrettyPrint.END_COLOUR}] {message}")

    @staticmethod
    def warn(message):
        print(f"[{PrettyPrint.WARN_ORANGE}WARN{PrettyPrint.END_COLOUR}] {message}")

    @staticmethod
    def fail(message):
        sys_exit(f"[{PrettyPrint.FAIL_RED}FAIL{PrettyPrint.END_COLOUR}] {message}")



# General imports
from dataclasses import dataclass
from enum import Enum
from typing import Optional, Dict
from distutils.version import StrictVersion
from functools import total_ordering


@total_ordering
class MacVersion(Enum):

    YOSEMITE = StrictVersion("10.10")

    EL_CAPITAN = StrictVersion("10.11")

    SIERRA = StrictVersion("10.12")

    HIGH_SIERRA = StrictVersion("10.13")

    MOJAVE = StrictVersion("10.14")

    CATALINA = StrictVersion("10.15")

    BIG_SUR = StrictVersion("11.0")

    MONTEREY = StrictVersion("12.0")

    VENTURA = StrictVersion("13.0")

    @staticmethod
    def from_version(version: str | StrictVersion):
        if version is None:
            raise ValueError("MacVersion: Can't identify none version number")

        if isinstance(version, str):
            version = StrictVersion(version)

        major_version = version.version[0]
        minor_version = version.version[1]

        if major_version < 11:
            truncated_version = StrictVersion(f"{major_version}.{minor_version}")
        else:
            truncated_version = StrictVersion(f"{major_version}.0")
        return MacVersion(truncated_version)

    @staticmethod
    def from_name(name: str) -> Optional['MacVersion']:
        if name is None:
            return None
        return MacVersion[name.upper()]

    def __lt__(self, other):
        return self.value < other.value


class VersionRange:

    def __init__(self, start: Optional[MacVersion] | Optional[StrictVersion] = None, end: Optional[MacVersion] | Optional[StrictVersion] = None):
        """
        A None start is taken to be the equivalent
        of "since the beginning of time"

        A None end is taken to be the equivalent of
        "currently supported"
        """

        if start is not None and not isinstance(start, StrictVersion) and not isinstance(start, MacVersion):
            raise ValueError(f"Unable to create a version range, passed in type for start was invalid [start={type(start)}], should have been a StrictVersion")

        if end is not None and not isinstance(end, StrictVersion) and not isinstance(start, MacVersion):
            raise ValueError(f"Unable to create a version range, passed in type for start was invalid [end={type(start)}], should have been a StrictVersion")

        self.start = start
        self.end = end

    def __str__(self):
        return f"VersionRange(start={self.start}, end={self.end})"

    def is_in_range(self, current_version: StrictVersion | MacVersion):
        # Infinite range case
        if self.start is None and self.end is None:
            return True

        if self.start is None and self.end is not None:
            return current_version <= self.end

        if self.start is not None and self.end is None:
            return self.start <= current_version

        # For somebody who likes to keep their books clean
        return self.start <= current_version <= self.end


@dataclass(frozen=True, order=True)
class DefaultSettings:
    name: str
    enabled: bool
    description: Optional[str]
    command: str
    start: Optional[MacVersion]
    end: Optional[MacVersion]
    check_command: Optional[str]
    expected_check_state: Optional[str]
    sudo: bool
    check_output: bool
    post_command: Optional[str]

    @staticmethod
    def from_dict(data: list[Dict]) -> list['DefaultSettings']:
        return [DefaultSettings.from_default_setting_dict(datum) for datum in data]

    @staticmethod
    def from_default_setting_dict(data: Dict) -> 'DefaultSettings':
        return DefaultSettings(
            name=data.get("name"),
            enabled=True if data.get("enabled") is None else data.get("enabled"),
            description=data.get("description"),
            command=data.get("command"),
            start=MacVersion.from_name(data.get("start")),
            end=MacVersion.from_name(data.get("end")),
            check_command=data.get("check_command"),
            expected_check_state=data.get("expected_check_state"),
            sudo=False if data.get("sudo") is None else data.get("sudo"),
            check_output=True if data.get("check_output") is None else data.get("check_output"),
            post_command=data.get("post_command"),
        )

    @property
    def version_range(self) -> VersionRange:
        return VersionRange(self.start, self.end)

    def should_run(self, current_version: StrictVersion | MacVersion) -> bool:
        if not self.enabled:
            return False

        if not self.version_range.is_in_range(current_version):
            return False

        if self.check_command is None:
            return True
        return True



# General imports
import logging
from pathlib import Path
import os
import json

# Project imports
from .enums import FileActionType
from .constants import PYDOTFILES_CACHE_DIRECTORY
from pydotfiles.utils import copy_file, symlink_file, rm_file, unsymlink_file, mv_file, run_file
from pydotfiles.utils import is_copied, is_linked


"""
Project-specific object primitives that
are used in higher-class business logic
"""

logger = logging.getLogger(__name__)


class FileAction:
    """
    Class representing an action on a file that we should
    have convenience methods for
    """

    def __init__(self, action, origin, destination, run_as_sudo, sudo_password=None):
        self.action = action
        self.origin = origin
        self.destination = destination
        self.run_as_sudo = run_as_sudo
        self.sudo_password = sudo_password

    def __str__(self):
        if self.action == FileActionType.SCRIPT or self.action == FileActionType.UNDO_SCRIPT:
            return f"{'SUDO ' if self.run_as_sudo else ''}{self.action.name} file={self.origin}, undo_file={self.destination}"
        return f"{'SUDO ' if self.run_as_sudo else ''}{self.action.name} {self.origin} -> {self.destination}"

    @property
    def is_completed(self):
        if self.action == FileActionType.COPY:
            return is_copied(self.origin, self.destination)
        elif self.action == FileActionType.SYMLINK:
            return is_linked(self.origin, self.destination)
        else:
            raise NotImplementedError(f"File Action: The action `{self.action}` is not supported yet (feel free to open a ticket on github!)")

    @property
    def reverse_action(self):
        return FileActionType.get_reverse(self.action)

    @property
    def destination_backup(self):
        return f"{self.destination}.backup"

    def do(self):
        logger.debug(f"File Action: Starting action [action={self.action}, origin={self.origin}, destination={self.destination}, use_sudo={self.run_as_sudo}]")
        if self.action == FileActionType.COPY:
            copy_file(self.origin, self.destination, self.run_as_sudo, self.sudo_password)
        elif self.action == FileActionType.SYMLINK:
            symlink_file(self.origin, self.destination, self.run_as_sudo, self.sudo_password)
        elif self.action == FileActionType.SCRIPT:
            run_file(self.origin, self.run_as_sudo, self.sudo_password)
        else:
            raise NotImplementedError(f"File Action: The action `{self.action}` is not supported yet (feel free to open a ticket on github!)")

        logger.debug(f"File Action: Successful action [action={self.action}, origin={self.origin}, destination={self.destination}, use_sudo={self.run_as_sudo}]")

    def undo(self):
        logger.debug(f"File Action: Starting undo action [action={self.reverse_action}, file={self.destination}, use_sudo={self.run_as_sudo}]")

        if self.destination is None:
            logger.info(f"File Action: No reverse action required [action={self.reverse_action}], file={self.destination}, use_sudo={self.run_as_sudo}")
            return

        if self.action == FileActionType.COPY:
            if not os.path.isfile(self.destination):
                logger.info(f"File Action: No reverse action required [action={self.reverse_action}], file={self.destination}, use_sudo={self.run_as_sudo}")
                return

            rm_file(self.destination, self.run_as_sudo, self.sudo_password)
        elif self.action == FileActionType.SYMLINK:
            if not os.path.islink(self.destination):
                logger.info(f"File Action: No reverse action required [action={self.reverse_action}], file={self.destination}, use_sudo={self.run_as_sudo}")
                return

            unsymlink_file(self.destination, self.run_as_sudo, self.sudo_password)
        elif self.action == FileActionType.SCRIPT:
            run_file(self.destination, self.run_as_sudo, self.sudo_password)
        else:
            raise NotImplementedError(f"File Action: The undo action `{self.action}` is not supported yet (feel free to open a ticket on github!)")

        logger.debug(f"File Action: Successful undo action [action={self.reverse_action}, file={self.destination}, use_sudo={self.run_as_sudo}]")

    def overwrite(self):
        # Just deletes the destination and then re-runs the operation
        logger.debug(f"File Action: Overwriting destination file [destination={self.destination}]")
        rm_file(self.destination, self.run_as_sudo, self.sudo_password)
        logger.debug(f"File Action: Successfully primed overwrite [destination={self.destination}]")

        self.do()

    def backup(self):
        logger.debug(f"File Action: Backing up file first [original={self.destination}, backup={self.destination_backup}]")
        mv_file(self.destination, self.destination_backup, self.run_as_sudo, self.sudo_password)
        logger.debug(f"File Action: Successfully backed up file first [original={self.destination}, backup={self.destination_backup}]")

        self.do()


class CacheDirectory:
    """
    Class representing the cache directory
    where we end up storing configurations
    related to the project
    """

    def __init__(self, package_manager=None, cache_directory=PYDOTFILES_CACHE_DIRECTORY):
        self.cache_directory = cache_directory
        self.package_manager = package_manager

        self.installed_packages = None
        self.installed_applications = None

        if package_manager is not None:
            self.reload_packages()
            self.reload_applications()

    @property
    def config_file(self):
        return f"{self.cache_directory}/config.json"

    @property
    def application_cache_file(self):
        return f"{self.cache_directory}/{self.package_manager.name.lower()}-application-cache"

    @property
    def package_cache_file(self):
        return f"{self.cache_directory}/{self.package_manager.name.lower()}-package-cache"

    @property
    def is_created(self):
        return os.path.isdir(self.cache_directory)

    """
    Config-file methods
    """

    def write_to_config(self, data):
        self.__idempotent_create__()

        with open(self.config_file, 'w') as config_file:
            config_file.write(json.dumps(data, sort_keys=True, indent=4))

    def read_from_config(self):
        if not os.path.isfile(self.config_file):
            return {}

        with open(self.config_file, 'r') as config_file:
            return json.load(config_file)

    """
    Public cache accessors
    """

    def is_package_installed(self, package):
        if self.installed_packages is None:
            return False

        return package in self.installed_packages

    def is_application_installed(self, application):
        if self.installed_applications is None:
            return False

        return application in self.installed_applications

    """
    Public cache updaters
    """

    def overwrite_packages(self, packages):
        self.__overwrite_cache_file__(self.package_cache_file, packages)

    def overwrite_applications(self, applications):
        self.__overwrite_cache_file__(self.application_cache_file, applications)

    def append_package(self, package):
        self.__append_to_cache_file__(self.package_cache_file, package)

    def append_application(self, application):
        self.__append_to_cache_file__(self.application_cache_file, application)

    def reload_packages(self):
        self.installed_packages = self.__read_from_cache_file__(self.package_cache_file)

    def reload_applications(self):
        self.installed_applications = self.__read_from_cache_file__(self.application_cache_file)

    """
    Internal helper methods
    """

    def __idempotent_create__(self):
        if self.is_created:
            logger.debug(f"Caching: Cache directory was already created [directory={self.cache_directory}]")
            return

        logger.debug(f"Caching: No cache directory was found, creating now [directory={self.cache_directory}]")
        Path(self.cache_directory).mkdir(parents=True, exist_ok=True)
        logger.debug(f"Caching: Successfully created cache directory [directory={self.cache_directory}]")

    def __overwrite_cache_file__(self, cache_file, data):
        self.__idempotent_create__()
        with open(cache_file, "w") as cache_file:
            cache_file.write(f"{data}\n")

    @staticmethod
    def __append_to_cache_file__(cache_file, data):
        with open(cache_file, "a") as cache_file:
            cache_file.write(f"{data}\n")

    def __read_from_cache_file__(self, cache_file):
        if os.path.isdir(self.cache_directory):
            if os.path.isfile(cache_file):
                with open(cache_file) as cache_file_descriptor:
                    cached_data = {line.rstrip('\n') for line in cache_file_descriptor}
                    logger.debug(f"Caching: Loaded in cache data [file={cache_file}, data={cached_data}")
                    return cached_data
            else:
                logger.debug(f"Caching: No cache file found [file={cache_file}]")
        else:
            logger.debug(f"Caching: No cache directory found [directory={self.cache_directory}]")
        return None



# General imports
from dataclasses import dataclass
from enum import Enum
from typing import Optional, Dict
from pathlib import Path

from .default_settings import DefaultSettings


class OSName(Enum):
    macos = "macos"
    linux = "linux"


class Shell(Enum):
    bash = "bash"
    zsh = "zsh"

    @staticmethod
    def get(name: Optional[str]) -> Optional['Shell']:
        if name is None:
            return None
        return Shell[name]


class PackageManager(Enum):
    brew = "brew"
    yum = "yum"
    apt = "apt"

    @staticmethod
    def get(name: Optional[str]) -> Optional['PackageManager']:
        if name is None:
            return None
        return PackageManager[name]


@dataclass(frozen=True, order=True)
class OperatingSystem:
    name: OSName
    shell: Optional[Shell]
    package_manager: Optional[PackageManager]
    packages: list[str]
    applications: list[str]
    default_dock: list[str]
    default_settings_files: list[Path]
    default_settings_data: list[DefaultSettings]

    @staticmethod
    def from_dict(data: Optional[Dict]) -> Optional['OperatingSystem']:
        if data is None:
            return None

        if data.get("default_settings_files") is None:
            default_settings_files = []
        else:
            default_settings_files = [Path(default_setting_file) for default_setting_file in data.get("default_settings_files")]

        return OperatingSystem(
            name=OSName[data.get("name")],
            shell=Shell.get(data.get("shell")),
            package_manager=PackageManager.get(data.get("package_manager")),
            packages=data.get("packages", []),
            applications=data.get("applications", []),
            default_dock=data.get("default_dock", []),
            default_settings_files=default_settings_files,
            default_settings_data=[],
        )

    def join_with_default_settings_data(self, additional_data: list[DefaultSettings]) -> 'OperatingSystem':
        return OperatingSystem(
            name=self.name,
            shell=self.shell,
            package_manager=self.package_manager,
            packages=self.packages,
            applications=self.applications,
            default_dock=self.default_dock,
            default_settings_files=self.default_settings_files,
            default_settings_data=additional_data,
        )

    def to_dict(self) -> Dict:
        return {
            "name": self.name.value,
            "shell": None if self.shell is None else self.shell.value,
            "package_manager": None if self.package_manager is None else self.package_manager.value,
            "packages": self.packages,
            "applications": self.applications,
            "default_dock": self.default_dock,
            "default_settings_files": [str(default_settings_file) for default_settings_file in self.default_settings_files],
        }



from setuptools import setup, find_packages
import pydotfiles

with open("README.md", "r") as readme_file:
    long_description = readme_file.read()

##
# If I ever come back to this and get confused about all the packaging stuff,
# just refer back to here: https://stackoverflow.com/a/58941536
##
setup(
    name="pydotfiles",
    version=pydotfiles.__version__,
    author="Jason Yao",
    author_email="Hello@JasonYao.com",
    description="Fast, easy, and automatic system configuration via a configuration file/repo",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/JasonYao/pydotfiles",
    packages=find_packages(exclude=("bin", "build", "dist", "git", "pydotfiles.egg-info")),
    include_package_data=True,
    package_data={
        '': [
            "**/*.json"
        ],
    },
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 3 - Alpha",

        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        "Topic :: System :: Systems Administration",

        # Supported python environments
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",

        # Supported operating systems
        "Operating System :: MacOS",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Unix",

        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    ],
    license="GPLv3",

    project_urls={
        'Bug Reports': 'https://github.com/JasonYao/pydotfiles/issues',
        # 'Funding': 'https://donate.pypi.org',
        # 'Say Thanks!': 'http://saythanks.io/to/example',
        'Source': 'https://github.com/JasonYao/pydotfiles',
    },

    # Installation dependencies below
    install_requires=[
        'PyYAML>=3.13',
        'jsonschema>=2.6.0',
        'GitPython>=2.1.11',
        'progressbar2>=3.38.0',
        'dataclasses;python_version<"3.7"',
        'importlib_resources; python_version < "3.9"',
    ],

    scripts=['pydotfiles/bin/pydotfiles'],
    zip_safe=False,
    python_requires='>=3.6',
    extra_require={
        'dev': [
            # Linters
            'autopep8',

            # Tests
            'pytest',
            'pytest-pep8',
            'pytest-cov',
        ],
        'release': [
            'wheel',
            'twine'
        ]
    }

)



import pytest
import uuid
from pydotfiles.models.validator import Validator
from pydotfiles.models.exceptions import ValidationError


"""
Validator tests
"""


def test_invalid_directory_none():
    # Setup
    validator = Validator()

    # System under test
    with pytest.raises(ValidationError):
        validator.validate_directory(None)


def test_invalid_directory_name():
    # Setup
    validator = Validator()
    invalid_directory_name = str(uuid.uuid4())

    # System under test
    with pytest.raises(ValidationError):
        validator.validate_directory(invalid_directory_name)


def test_valid_directory(tmpdir):
    # Setup
    validator = Validator()
    valid_file = tmpdir.join("settings.json")
    valid_file_content = "{\"version\": \"alpha\", \"schema\": \"core\"}"
    valid_file.write(valid_file_content)

    # System under test
    validator.validate_directory(tmpdir.strpath)


def test_valid_directory_multiple_modules(tmpdir):
    # Setup
    validator = Validator()
    modules_settings_file_a = tmpdir.mkdir("a_directory").join("settings.json")
    modules_settings_file_a_content = "{\"version\": \"alpha\", \"schema\": \"core\"}"
    modules_settings_file_a.write(modules_settings_file_a_content)

    modules_settings_file_b = tmpdir.mkdir("b_directory").join("settings.yaml")
    modules_settings_file_b_content = "version: \"alpha\"\nschema: \"core\""
    modules_settings_file_b.write(modules_settings_file_b_content)

    # System under test
    validator.validate_directory(tmpdir.strpath)


def test_invalid_file_none():
    # Setup
    validator = Validator()

    # System under test
    with pytest.raises(ValidationError):
        validator.validate_file(None)


def test_invalid_file_name():
    # Setup
    validator = Validator()
    invalid_file_name = str(uuid.uuid4())

    # System under test
    with pytest.raises(ValidationError):
        validator.validate_file(invalid_file_name)


def test_invalid_file_empty_file(tmpdir):
    # Setup
    validator = Validator()
    invalid_file = tmpdir.join("invalid_json.json")
    invalid_file_content = ""
    invalid_file.write(invalid_file_content)

    # System under test
    with pytest.raises(ValidationError):
        validator.validate_file(invalid_file.strpath)


def test_invalid_file_invalid_json(tmpdir):
    # Setup
    validator = Validator()
    invalid_file = tmpdir.join("invalid_json.json")
    invalid_file_content = "{\"version\": \"alpha,}"
    invalid_file.write(invalid_file_content)

    # System under test
    with pytest.raises(ValidationError):
        validator.validate_file(invalid_file.strpath)


def test_valid_file(tmpdir):
    # Setup
    validator = Validator()
    valid_file = tmpdir.join("valid_json.json")
    valid_file_content = "{\"version\": \"alpha\", \"schema\": \"core\"}"
    valid_file.write(valid_file_content)

    # System under test
    validator.validate_file(valid_file.strpath)


def test_invalid_data_none():
    # Setup
    validator = Validator()

    # System under test
    with pytest.raises(ValidationError):
        validator.validate_data(None)
